# This file contains a set of helper/utility routines for improving the
# notebook calculations
import numpy as np
from typing import Iterable, Optional


def calc_harmonic_work_theory(
    velocity: float, trap_strength: float, time: Iterable,
    beta: Optional[float] = 1.0, D: Optional[float] = 1.0
) -> Iterable:
    prefactor = (velocity ** 2) / (beta * D)
    numer = (1 - np.exp(-beta * D * trap_strength * np.array(time)))
    denom = beta * D * trap_strength

    return prefactor * (np.array(time) - (numer / denom))


def calc_harmonic_power_theory(
    velocity: float, trap_strength: float, time: Iterable,
    beta: Optional[float] = 1.0, D: Optional[float] = 1.0
) -> Iterable:
    prefactor = (velocity ** 2) / (beta * D)
    argument = 1 - np.exp(-beta * D * trap_strength * np.array(time))
    return prefactor * argument



import os
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Optional, Iterable

from fokker_planck.util import notebooks


# SECTION - Helper routines
def _configure_visualization_format():
    sns.set(style="darkgrid")


def _check_input_file_specs(write_name: str, write_path: str) -> bool:
    if write_name is not None or write_path is not None:
        if write_name is None or write_path is None:
            raise ValueError(
                "You must supply BOTH `write_path` and `write_name`"
            )
        return True
    else:
        return False

# !SECTION

# SECTION - General routines (multi-notebook)


def density_tracking_plot(
    density_tracker: List, time_tracker: List, norm_tracker: List,
    x_array: np.ndarray, write_name: str, write_path: str,
    write_format: Optional[str] = "pdf",
) -> plt.Axes:

    save_file = _check_input_file_specs(write_name, write_path)
    fig, ax = plt.subplots(2, 1, figsize=(6.3, 4.5))
    Pal = sns.color_palette("Spectral", len(density_tracker))

    for i, prob in enumerate(density_tracker):
        ax[0].plot(x_array, prob, linewidth=2.5, color=Pal[i])

    ax[1].plot(time_tracker, norm_tracker, linewidth=2.5, color='k')

    ax[0].set_xlabel(r"Position $x$", fontsize=15)
    ax[0].set_ylabel(r"$p(x)$", fontsize=15)

    ax[1].set_xlabel(r"Time $t$", fontsize=15)
    ax[1].set_ylabel(r"$\sum p(x_i)\Delta x_i$", fontsize=15)

    ax[1].set_ylim([0, 1.1])

    plt.tight_layout()
    if save_file:
        plt.savefig(os.path.join(write_path, write_name), format=write_format)
    return fig

# !SECTION

# SECTION - Notebook 1 routines - Advection


# !SECTION

# SECTION - Notebook 2 routines - Diffusion
# !SECTION

# SECTION - Notebook 3 routines - Advection Diffusion - relaxation in harmonic BCs

def mean_tracking_plot_harmonic(
    trap_strength: float, D: float, mean_tracker: Iterable[float],
    time_tracker: Iterable[float], write_name: Optional[str] = None,
    write_path: Optional[str] = None, write_format: Optional[str] = "pdf",
) -> plt.Axes:
    _configure_visualization_format()
    save_file = _check_input_file_specs(write_name, write_path)

    fig, ax = plt.subplots(1, 1, figsize=(6.3, 3.5))
    
    time_theory = np.linspace(0, time_tracker[-1], 100)
    relax_theory = np.exp(-trap_strength * time_theory / D)

    mean_tracker /= mean_tracker[0]
    ax.fill_between(time_theory, relax_theory, color=sns.xkcd_rgb["tomato"], alpha=0.2)
    ax.plot(
        time_theory, relax_theory, linewidth=2.5,
        color=sns.xkcd_rgb["tomato"], label="Theory"
    )
    ax.plot(
        time_tracker, mean_tracker, 'o',
        markersize=7, color=sns.xkcd_rgb["electric blue"],
        alpha=0.8, label="Simulation"
    )

    ax.legend(fontsize=12)
    ax.set_xlabel(r"Elapsed Time", fontsize=15)
    ax.set_ylabel(r"Distribution mean $\langle x\rangle$", fontsize=15)
    plt.tight_layout()
    if save_file:
        plt.savefig(os.path.join(write_path, write_name), format=write_format)

    return fig


def harmonic_work_plot(
    work_arr: Iterable, power_arr: Iterable, time: Iterable, vel_arr: Iterable,
    vel_labels: Iterable, trap_strength: float,
    write_dir: Optional[str] = None, write_name: Optional[str] = None
) -> plt.Axes:
    _configure_visualization_format()
    save_file = _check_input_file_specs

    fig, ax = plt.subplots(1, 2, figsize=(6.3, 3.5))
    Pal = sns.color_palette("husl", len(vel_arr))

    theory_arr_work = [
        notebooks.calc_harmonic_power_theory(v, trap_strength, time)
        for v in vel_arr
    ]

    theory_arr_power = [
        notebooks.calc_harmonic_power_theory(v, trap_strength, time)
        for v in vel_arr
    ]

    # Generate work plots
    for i, (w, w_t) in enumerate(zip(work_arr, theory_arr_work)):
        ax[0].plot(
            time, w_t, '--', linewidth=1.0, color='k', alpha=0.6,
            label=(lambda x: "Theory" if x == 0 else None)(i)
        )
        ax[0].plot(time, w, linewidth=2.0, color=Pal[i], label=vel_labels[i])

    # Generate power plots
    for i, (p, p_t) in enumerate(zip(power_arr, theory_arr_power)):
        ax[1].plot(time, p, linewidth=2.0, color=Pal[i])
        ax[1].plot(time, p_t, '--', linewidth=1.0, color='k', alpha=0.6)

    ax[0].legend(fontsize=12)
    ax[0].set_xlabel(r"Elapsed time $\tau$", fontsize=15)
    ax[0].set_ylabel(r"Mean work $\langle W\rangle_{[0, \tau]}$", fontsize=15)
    ax[1].set_xlabel(r"Elapsed time $\tau$", fontsize=15)
    ax[1].set_ylabel(r"Mean power input $\langle P(\tau)\rangle$", fontsize=15)

    plt.tight_layout()
    if save_file:
        plt.savefig(os.path.join(write_dir, write_name), format="pdf")
    return fig


# !SECTION


# SECTION - Notebook 4 routines - Excess work and Power - Theory and numerical
# !SECTION

# SECTION Notebook 5 routines - Periodic system and flux
# !SECTION

# SECTION Notebook 6 routines - Erasure Protocol?
# !SECTION






from typing import Union, Iterable
from enum import Enum, auto


class DiffScheme(Enum):
    CrankNicolson = auto()
    Implicit = auto()
    Explicit = auto()


class BoundaryCondition(Enum):
    HardWall = auto()
    Periodic = auto()
    Open = auto()


class SplitMethod(Enum):
    Lie = auto()
    Strang = auto()
    SymStrang = auto()


CPVector = Union[float, Iterable]



# This is a class-based implementation of a simulator object that can
# run simulations of particular FPE instances. Effectively this acts as
# a convenience wrapper around the raw fpe integraator obects
from typing import Iterable, Optional, Dict, Callable
import numpy as np

from fokker_planck.types.basetypes import CPVector
from fokker_planck.simulator.base import StaticSimulator1D, DynamicSimulator1D
import fokker_planck.forceFunctions as ff


class Protocol:
    def __init__(self, control_parameter_vals: Iterable, time_vals: Iterable):
        self._CP = control_parameter_vals
        self._time = time_vals

    @property
    def CP(self):
        return self._CP

    @property
    def time(self):
        return self._time


# STATIC SIMUlATORS
class HarmonicEquilibrationSimulator(StaticSimulator1D):
    def __init__(
        self, fpe_config: Dict, k_trap: float, trap_min: Optional[float] = 0
    ):
        super().__init__(fpe_config)
        self.force_func = ff.harmonic_force
        self.force_params = [k_trap, trap_min]

    def initialize_probability(self, init_var: Optional[float] = None):
        if init_var is None:
            x_len = len(self.fpe_args.x_array)
            uni_prob = (np.ones(x_len) / ((x_len - 2) * self.fpe_args.dx))
            uni_prob[0] = 0
            uni_prob[-1] = 0
            self.fpe_prob = uni_prob
        else:
            self.fpe.initialize_probability(self.force_params[1], init_var)


class PeriodicEquilibrationSimulator(StaticSimulator1D):
    def __init__(
        self, fpe_config: Dict, amp: float, n_minima: int,
        phase_shift: float = 0
    ):
        super().__init__(fpe_config)
        self.force_func = ff.periodic_force
        self.force_params = [amp, n_minima, phase_shift]

    def initialize_probability(self, classification: Optional[str] = None, **kwargs):
        if classification is None:
            x_len = len(self.fpe_args.x_array)
            uni_prob = np.ones(x_len) / (x_len * self.fpe_args.dx)
            self.fpe_prob = uni_prob

        if classification.lower() == 'gaussian':
            self.fpe.initialize_probability(**kwargs)

        raise NotImplementedError(
            f'Probability classificaition {classification} not recognized, '
            'currently only None (uniform) and gaussian are supported'
        )


# DYNAMIC SIMULATORS
class BreathingSimulator(DynamicSimulator1D):
    def __init__(
        self, fpe_config: Dict, k_i: float, k_f: float,
        forceFunction: Callable, energyFunction: Callable
    ):
        super().__init__(fpe_config, k_i, k_f)
        self.forceFunc = forceFunction
        self.energyFunc = energyFunction

    def build_friction_array(self) -> np.ndarray:
        return self.lambda_array ** (3/2)

    def initialize_probability(self):
        self.fpe.initialize_probability(0, 1 / self.lambda_init)

    def update(self, protocol_bkw: CPVector, protocol_fwd: CPVector):
        params_bkw = ([protocol_bkw, 0])
        params_fwd = ([protocol_fwd, 0])
        if not self.check_cfl(params_fwd, self.force_func):
            raise ValueError('CFL not satisfied!')

        self.fpe.work_step(
            params_bkw, params_fwd, self.forceFunc, self.energyFunc
        )


class HarmonicTranslationSimulator(DynamicSimulator1D):
    def __init__(
        self, fpe_config: Dict, trap_init: float,
        trap_fin: float, trap_strength: float,
        force_function: Optional[Callable] = ff.harmonic_force_const_velocity,
        energy_function: Optional[Callable] = ff.harmonic_energy_const_velocity,
    ):
        super().__init__(fpe_config, trap_init, trap_fin)
        self.force_func = force_function
        self.energy_func = energy_function
        self.trap_strength = trap_strength

    def build_friction_array(self) -> np.ndarray:
        return np.ones_like(self.lambda_array)

    def initialize_probability(self):
        self.fpe.initialize_probability(self.lambda_init, 1 / self.trap_strength)

    def update(self, protocol_bkw: CPVector, protocol_fwd: CPVector):
        dlambda = protocol_fwd - protocol_bkw
        cp_vel = dlambda / self.fpe.dt
        # Params: [k_trap, center, velocity, D]
        params_bkw = ([self.trap_strength, 0, cp_vel, self.fpe.D])
        params_fwd = ([self.trap_strength, dlambda, cp_vel, self.fpe.D])

        if not self.check_cfl(params_fwd, self.force_func):
            raise ValueError('CFL not satisfied!')

        self.fpe.work_step(
            params_bkw, params_fwd, self.force_func, self.energy_func
        )


if __name__ == "__main__":
    # 2 equivalent sample input_config dictionaries
    config_1 = {
        "D": 1.0,
        "dx": 0.01,
        "dt": 0.001,
        "x_min": -3,
        "x_max": 3
    }

    config_2 = {
        "D": 1.0,
        "dx": 0.01,
        "dt": 0.001,
        "x_array": np.arange(-3, 3, 0.01)
    }

    breathing_1 = BreathingSimulator(
        config_1, 0.5, 4.0, ff.harmonic_force, ff.harmonic_energy
    )
    breathing_2 = BreathingSimulator(
        config_2, 0.5, 4.0, ff.harmonic_force, ff.harmonic_energy
    )

    harmonic_trap = HarmonicTranslationSimulator(
        config_1, 0, 5, 8, ff.harmonic_force_const_velocity,
        ff.harmonic_energy_const_velocity
    )

    proto_n_1 = breathing_1.build_protocol(1.5, mode="naive")
    proto_n_2 = breathing_2.build_protocol(1.5, mode="naive")

    proto_o_1 = breathing_1.build_protocol(1.5, mode="optimal")
    proto_o_2 = breathing_2.build_protocol(1.5, mode="optimal")

    proto_n_3 = harmonic_trap.build_protocol(1.5, mode="naive")
    proto_n_4 = harmonic_trap.build_protocol(3.0, mode="naive")

    proto_o_3 = harmonic_trap.build_protocol(1.5, mode="optimal")
    proto_o_4 = harmonic_trap.build_protocol(3.0, mode="optimal")

    print("DONE!")



from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum, auto
import numpy as np
import scipy.interpolate
from typing import Dict, Optional, Union, Iterable, Tuple, Callable

from fokker_planck.types.basetypes import CPVector
from fokker_planck.integrator import FokkerPlanck1D
from fokker_planck.base import Integrator


class BaseSimulator(ABC):

    def __init__(self, tracking_stride: Optional[int] = 10):
        self.tau = None
        self.time_array = None
        self.tracking_stride = tracking_stride

        self.force_func: Callable = None
        self.fpe: Integrator = None

    def check_cfl(self, params):
        if not self.fpe.check_CFL(params, self.force_func):
            raise ValueError('CFL not satisfied!')


class KeyElements1D(Enum):
    D: float = auto()
    dx: float = auto()
    dt: float = auto()
    x_min: float = auto()
    x_max: float = auto()
    x_array: float = auto()

    @classmethod
    def __iter__(cls):
        return iter(cls._member_map_.values())

    @classmethod
    def members(cls):
        return cls._member_names_


class InputParams:
    def __init__(self, params: Optional[Dict] = None):
        if params is not None:
            for k, p in params.items():
                if p is None:
                    continue
                self.__dict__[k] = p

    @property
    def as_dict(self):
        return vars(self)


class Simulator1D(BaseSimulator):

    def __init__(self, fpe_config: Dict, tracking_stride: int = 10):
        super().__init__(tracking_stride=tracking_stride)
        self.key_elems = KeyElements1D
        self.fpe_args, self.fpe_kwargs = self._parse_input_config(fpe_config)
        self.fpe = FokkerPlanck1D(
            **self.fpe_args.as_dict, **self.fpe_kwargs.as_dict
        )

    def reset(self):
        self.fpe.reset()

    def _parse_input_config(
        self, input_config: Dict
    ) -> Tuple[InputParams, InputParams]:
        # Pull of key attributes
        key_elem_dict = defaultdict(lambda: None)
        # for k in key_elements:
        for k in self.key_elems:
            key_elem_dict[k.name] = input_config.get(k.name, None)

        if (
            (key_elem_dict['x_array'] is None)
            and any([key_elem_dict[x] is None for x in ['x_min', 'x_max']])
        ):
            raise ValueError(
                'Must provide either `x_array` or `x_min` AND `x_max` in'
                'input_config dictionary'
            )

        if any([key_elem_dict[x] is None for x in ['D', 'dx', 'dt']]):
            raise ValueError(
                'Must provide keys of `D`, `dx`, and `dt` in input config '
                'dictionary.'
            )

        if key_elem_dict['x_array'] is None:
            x_min, x_max, dx = key_elem_dict['x_min'], key_elem_dict['x_max'], key_elem_dict['dx']
            key_elem_dict['x_array'] = np.arange(x_min, x_max, dx)
            del key_elem_dict['x_min']
            del key_elem_dict['x_max']

        additional_specs = {
            k: v for k, v in input_config.items() if k not in self.key_elems.members()
        }
        return InputParams(key_elem_dict), InputParams(additional_specs)


class StaticSimulator1D(Simulator1D):
    def __init__(self, fpe_config: Dict, tracking_stride: int = 10):
        super().__init__(fpe_config, tracking_stride=tracking_stride)
        self.force_params: Iterable = None
        self.force_func: Callable = None

    def run_simulation(self, tau: float):

        if self.force_params is None or self.force_func is None:
            raise ValueError(
                "Must initialize/provide `force_params` and `force_func` "
                "instance variables in StaticSimulator1D subclasses"
            )

        prob_tracker = []
        time_tracker = []
        time_array = np.arange(0, tau, self.fpe_args.dt)

        self.initialize_probability()
        self.fpe.init_physical_trackers()

        self.check_cfl(self.force_params)

        for i, t in enumerate(time_array):
            if i % self.tracking_stride:
                prob_tracker.append(self.fpe.prob)
                time_tracker.append(t)
            self.fpe.integrate_step(self.force_params, self.force_func)

        return SimulationResult(
            self, prob_tracker, time_tracker, sim_type='static'
        )

    @abstractmethod
    def initialize_probability(self):
        pass


class DynamicSimulator1D(Simulator1D):

    def __init__(
        self, fpe_config: Dict, lambda_init: CPVector, lambda_fin: CPVector,
        n_lambda: Optional[int] = 500, tracking_stride: int = 10
    ):
        super().__init__(fpe_config, tracking_stride=tracking_stride)
        self.lambda_array = np.linspace(lambda_init, lambda_fin, n_lambda)
        self.lambda_init = lambda_init
        self.lambda_fin = lambda_fin

    def _get_raw_velocity(self, friction: Union[float, Iterable]):
        return 1.0 / np.sqrt(friction)

    def _get_raw_times(self, raw_velocity: np.ndarray) -> np.ndarray:
        raw_times = np.zeros_like(self.lambda_array)
        for i in range(len(raw_times) - 1):
            numerator = 2 * (self.lambda_array[i+1] - self.lambda_array[i])
            denominator = (raw_velocity[i+1] + raw_velocity[i])
            raw_times[i+1] = (numerator / denominator) + raw_times[i]

        return raw_times

    def _get_real_times(self, raw_times: np.ndarray, tau: float) -> np.ndarray:
        raw_tau = raw_times[-1]
        return (tau / raw_tau) * raw_times

    def _get_real_velocities(
        self, raw_times: np.ndarray, raw_velocities: np.ndarray, tau: float
    ) -> np.ndarray:
        raw_tau = raw_times[-1]
        return (raw_tau / tau) * raw_velocities

    def _build_optimal_protocol(self, tau: float) -> np.ndarray:
        raw_velocity = self._get_raw_velocity(self.build_friction_array())
        raw_times = self._get_raw_times(raw_velocity)
        self.real_time = self._get_real_times(raw_times, tau)
        self.real_velocity = self._get_real_velocities(raw_times, raw_velocity, tau)

        # Now, we can generate the pairs of (time, k_value) from the pairs of real_times, k_array
        # But we want a uniform set of times at which we know the k values, so we cna interpolate
        time_interp = scipy.interpolate.interp1d(self.real_time, self.lambda_array, fill_value="extrapolate")
        self.optimal_protocol = time_interp(self.time_array)

    def _build_naive_protocol(self):
        self.naive_protocol = np.linspace(self.lambda_init, self.lambda_fin, len(self.time_array))

    def build_protocol(self, tau: float, mode: Optional[str] = "naive"):
        n_steps = int(tau / self.fpe.dt)
        self.time_array = np.linspace(0, tau, n_steps)

        if mode == "naive":
            self._build_naive_protocol()
            protocol = self.naive_protocol
        elif mode == "optimal":
            self._build_optimal_protocol(tau)
            protocol = self.optimal_protocol
        else:
            raise ValueError("`mode` parameter must be `naive` or `optimal`")

        return protocol, self.time_array

    def run_simulation(self, tau: float, mode: Optional[str] = "naive"):

        protocol, _ = self.build_protocol(tau, mode=mode)

        prob_tracker = []
        time_tracker = []
        cp_tracker = []

        self.initialize_probability()
        self.fpe.init_physical_trackers()

        for i, p in enumerate(protocol[:-1]):
            if i % 10 == 0:
                prob_tracker.append(self.fpe.prob)
                time_tracker.append(self.time_array[i])
                cp_tracker.append(p)

            self.update(p, protocol[i+1])

        return SimulationResult(
            self, prob_tracker, time_tracker, sim_type='dynamic',
            cp_tracker=cp_tracker
        )

    @abstractmethod
    def build_friction_array(self) -> np.ndarray:
        pass

    @abstractmethod
    def initialize_probability(self):
        pass

    @abstractmethod
    def update(self, protocol_bkw: CPVector, protocol_fwd: CPVector):
        pass


class SimulationResult:
    def __init__(
        self, sim: BaseSimulator, prob_tracker: Iterable,
        time_tracker: Iterable, sim_type: str,
        cp_tracker: Optional[Iterable] = None
    ):
        self._sim = sim
        self.CP = cp_tracker
        self.prob_tracker = prob_tracker
        self.time = time_tracker
        self.sim_type = sim_type

    # Helper classes?


# Empty class implementation for testing base functionality
class VoidSimulator(DynamicSimulator1D):
    def __init__(self, fpe_config: Dict):
        super().__init__(fpe_config, 0, 1)

    def build_friction_array(self):
        pass

    def initialize_probability(self):
        pass

    def update(self):
        pass



from typing import Callable, Optional, Tuple, Dict
import numpy as np
import scipy
import scipy.sparse as sp
from scipy.sparse.linalg import inv
from fokker_planck.base import Integrator


# ANCHOR 2D Integrator
class FPE_integrator_2D(Integrator):

    def __init__(
        self, D: float, dt: float, dx: float, dy: float, xArray: float,
        yArray: float, diffScheme: Optional[str] = 'crank-nicolson',
        adScheme: Optional[str] = 'lax-wendroff',
        boundaryCond: Optional[str] = 'hard-wall',
        splitMethod: Optional[str] = 'strang',
        output: Optional[bool] = False,
        constDiff: Optional[bool] = True
    ):

        super().__init__(
            D, dt, diffScheme, adScheme, boundaryCond, splitMethod, output,
            constDiff
        )
        self.dx = dx
        self.dy = dy
 
        self.Nx = len(xArray)
        self.Ny = len(yArray)
        self.N = self.Nx * self.Ny

        # self.prob = np.ones((self.N,self.N))/(sum(self.Nx*self.dx)*sum(self.Ny*self.dy))
        self.prob = np.ones(self.Nx * self.Ny) / (self.Nx * self.Ny * self.dx * self.dy)
        self.xArray = xArray
        self.yArray = yArray

        self.constDiff = constDiff
        self.sparTest = False

        self.init_diffusion_matrix()

    @property
    def dimension(self) -> int:
        return 2

    @property
    def mean(self) -> Dict:
        prob_grid = self.prob.reshape(self.Nx, self.Ny)
        marginal_x = np.sum(prob_grid, axis=1) * self.dy
        marginal_y = np.sum(prob_grid, axis=0) * self.dx

        mean_x = np.sum(marginal_x * self.dx * self.xArray)
        mean_y = np.sum(marginal_y * self.dy * self.yArray)

        return {'x_mean': mean_x, 'y_mean': mean_y}

    @property
    def covariance(self) -> np.ndarray:
        pass

    def _flatten_probability(self, prob: np.ndarray):
        pass

    def reset(
        self, covariance: Optional[np.ndarray] = None,
        mean: Optional[np.ndarray] = None
    ):
        if covariance is not None and mean is not None:
            self.initializeProbability(mean, covariance)
        else:
            self.prob = np.ones((self.Nx, self.Ny)) / (self.Nx * self.Ny * self.dx * self.dy)
        self.init_physical_trackers()

    def init_physical_trackers(self):
        self.workAccumulator = 0
        self.workTracker = []
        self.powerTracker = []
        self.timeTracker = []

        self.xFlux = np.zeros(self.Nx)
        self.yFlux = np.zeros(self.Ny)
        self.xFluxTracker = 0
        self.yFluxTracker = 0

    def initializeProbability(self, mean: np.ndarray, covariance: np.ndarray):
        pass

    def initializeUserProbability(
        self, func: Callable, params: Optional[Tuple] = None
    ):
        pass

    def init_diffusion_matrix(self):
        if(self.output is True):
            print("\n\nInitializing diffusion term integration matrix...\n")

        self._setDiffusionScheme()

        if(self.output is True):
            print("\t\tInitializing integration matrices for diffusion\n")

        alpha = self.D * self.dt / (self.dx * self.dx)

        self.AMat = sp.lil_matrix((self.N, self.N))
        self.BMat = sp.lil_matrix((self.N, self.N))

        # Bulk term initializations
        self.AMat.setdiag(1 + 4 * alpha * self.exp_imp)
        self.AMat.setdiag(-1 * self.exp_imp * alpha, k=1)
        self.AMat.setdiag(-1 * self.exp_imp * alpha, k=-1)
        self.AMat.setdiag(-1 * self.exp_imp * alpha, k=self.Nx)
        self.AMat.setdiag(-1 * self.exp_imp * alpha, k=-self.Nx)

        self.BMat.setdiag(1 - 4 * alpha * (1 - self.exp_imp))
        self.BMat.setdiag(alpha * (1 - self.exp_imp), k=1)
        self.BMat.setdiag(alpha * (1 - self.exp_imp), k=-1)
        self.BMat.setdiag(alpha * (1 - self.exp_imp), k=self.Nx)
        self.BMat.setdiag(alpha * (1 - self.exp_imp), k=-self.Nx)

        self._initializeBoundaryTerms(alpha)

        # convert to csc format
        self.AMat = self.AMat.tocsc()
        self.BMat = self.BMat.tocsc()

        self.CMat = inv(self.AMat).dot(self.BMat)

    def _initializeBoundaryTerms(self, alpha: float):
        self._matrixBoundary_A(alpha, 0)
        self._matrixBoundary_B(alpha, 0)

        self._matrixBoundary_A(alpha, self.Nx - 1)
        self._matrixBoundary_B(alpha, self.Nx - 1)

    def _matrixBoundary_A(self, alpha: float, idx: int):
        if self.BC == 'periodic':
            # x-dimension BCs

            # Need to place top right and bottom left elements in each block
            # matrix for X and Y directions
            for image in range(0, self.Ny):
                x_idx = idx + self.Nx * image
                y_idx_fw = (idx + 1) % self.Nx + self.Nx * image
                y_idx_rv = (idx - 1) % self.Nx + self.Nx * image

                self.AMat[x_idx, y_idx_fw] = -self.exp_imp * alpha
                self.AMat[x_idx, y_idx_rv] = -self.exp_imp * alpha

            for image in range(1, self.Ny):
                self.AMat[image * self.Nx - 1, image * self.Nx] = 0
                self.AMat[image * self.Nx, image * self.Nx - 1] = 0

            # y-dimension BCs
            self.AMat.setdiag(-self.exp_imp * alpha, k=self.N - self.Nx)
            self.AMat.setdiag(-self.exp_imp * alpha, k=-(self.N - self.Nx))

        elif self.BC == "hard-wall":
            # This resolves the HW conditions in the X-dimension, still need
            # to resolve in y
            # NOTE Need to update this, I aded the 8, but I need to figure out
            # where the 4 goes and make sure the 2s are right, as well as
            # resolve the multi-coordinate boundaries,..

            # Set necessary elemetns to zero
            for image in range(1, self.Ny):
                self.AMat[image * self.Nx - 1, image * self.Nx] = 0
                self.AMat[image * self.Nx, image * self.Nx - 1] = 0

            # NOTE THIS IS INCORRECT....
            # X-boundaries only
            # Loop over all y values corresponding to the X at idx, which is
            # y = 1 --> N_y - 2
            for image in range(1, self.Ny - 1):
                row_idx_top = image * self.Nx
                row_idx_bot = (image + 1) * self.Nx - 1

                self.AMat[row_idx_top, row_idx_top] = 1 + 4 * alpha
                self.AMat[row_idx_bot, row_idx_bot] = 1 + 4 * alpha

                self.AMat[row_idx_top, row_idx_top + 1] = -2 * alpha
                self.AMat[row_idx_top, row_idx_top + self.Nx] = -1 * alpha
                self.AMat[row_idx_top, row_idx_top - self.Nx] = -1 * alpha

                self.AMat[row_idx_bot, row_idx_bot - 1] = -2 * alpha
                self.AMat[row_idx_bot, row_idx_bot + self.Nx] = -1 * alpha
                self.AMat[row_idx_bot, row_idx_bot - self.Nx] = -1 * alpha

            # Y-boundaries only
            for image in range(1, self.Nx - 1):
                row_idx_top = image
                row_idx_bot = image + (self.Ny - 1) * self.Nx

                self.AMat[row_idx_top, row_idx_top] = 1 + 4 * alpha
                self.AMat[row_idx_bot, row_idx_bot] = 1 + 4 * alpha

                self.AMat[row_idx_top, row_idx_top + 1] = -1 * alpha
                self.AMat[row_idx_top, row_idx_top - 1] = -1 * alpha
                self.AMat[row_idx_top, row_idx_top + self.Nx] = -2 * alpha

                self.AMat[row_idx_bot, row_idx_bot + 1] = -1 * alpha
                self.AMat[row_idx_bot, row_idx_bot - 1] = -1 * alpha
                self.AMat[row_idx_bot, row_idx_bot - self.Nx] = -2 * alpha

            # Corner boundaries
            row_idx_00 = 0
            row_idx_N0 = self.Nx - 1
            row_idx_0N = (self.Ny - 1) * self.Nx
            row_idx_NN = (self.Ny * self.Nx) - 1

            self.AMat[row_idx_00, row_idx_00] = 1 + 4 * alpha
            self.AMat[row_idx_00, row_idx_00 + 1] = -2 * alpha
            self.AMat[row_idx_00, row_idx_00 + self.Nx] = -2 * alpha

            self.AMat[row_idx_N0, row_idx_N0] = 1 + 4 * alpha
            self.AMat[row_idx_N0, row_idx_N0 - 1] = -2 * alpha
            self.AMat[row_idx_N0, row_idx_N0 + self.Nx] = -2 * alpha

            self.AMat[row_idx_0N, row_idx_0N] = 1 + 4 * alpha
            self.AMat[row_idx_0N, row_idx_0N + 1] = -2 * alpha
            self.AMat[row_idx_0N, row_idx_0N - self.Nx] = -2 * alpha

            self.AMat[row_idx_NN, row_idx_NN] = 1 + 4 * alpha
            self.AMat[row_idx_NN, row_idx_NN - 1] = -2 * alpha
            self.AMat[row_idx_NN, row_idx_NN - self.Nx] = -2 * alpha

        elif self.BC == "open":
            for image in range(1, self.Ny):
                self.AMat[image * self.Nx - 1, image * self.Nx] = 0
                self.AMat[image * self.Nx, image * self.Nx - 1] = 0

        else:
            raise ValueError(
                f"Invalid boundary condition `{self.BC}`, cannot resolve "
                "diffusion matrix A"
            )

    def _matrixBoundary_B(self, alpha: float, idx: int):

        if self.BC == "periodic":
            # x-dimension BCs
            for image in range(0, self.Ny):
                x_idx = idx + self.Nx * image
                y_idx_fw = (idx + 1) % self.Nx + self.Nx * image
                y_idx_rv = (idx - 1) % self.Nx + self.Nx * image

                self.BMat[x_idx, y_idx_fw] = alpha * (1 - self.exp_imp)
                self.BMat[x_idx, y_idx_rv] = alpha * (1 - self.exp_imp)

            for image in range(1, self.Ny):
                self.BMat[image * self.Nx - 1, image * self.Nx] = 0
                self.BMat[image * self.Nx, image * self.Nx - 1] = 0

            # y-dimension BCs
            self.BMat.setdiag(alpha * (1 - self.exp_imp), k=self.N - self.Nx)
            self.BMat.setdiag(alpha * (1 - self.exp_imp), k=-(self.N - self.Nx))

        elif self.BC == "hard-wall":
            # for image in range(0, self.Ny):
            #     _idx = idx + self.Nx * image
            #     self.BMat[_idx, _idx] = 1
            #     self.BMat[_idx, abs(idx - 1) + self.Nx * image] = 0

            for image in range(1, self.Ny):
                self.BMat[image * self.Nx - 1, image * self.Nx] = 0
                self.BMat[image * self.Nx, image * self.Nx - 1] = 0

            # for image in range(1, self.Ny):
            #     self.BMat[image * self.Nx - 1, image * self.Nx] = 0
            #     self.BMat[image * self.Nx, image * self.Nx - 1] = 0

            # for diag_idx in range(self.Nx):
            #     self.BMat[diag_idx, self.Nx + diag_idx] = 0
            #     self.BMat[self.N - 1 - diag_idx, self.N - self.Nx - 1 - diag_idx] = 0

            for image in range(1, self.Ny - 1):
                row_idx_top = image * self.Nx
                row_idx_bot = (image + 1) * self.Nx - 1

                self.BMat[row_idx_top, row_idx_top] = 1
                self.BMat[row_idx_bot, row_idx_bot] = 1

                self.BMat[row_idx_top, row_idx_top + 1] = 0
                self.BMat[row_idx_top, row_idx_top + self.Nx] = 0
                self.BMat[row_idx_top, row_idx_top - self.Nx] = 0

                self.BMat[row_idx_bot, row_idx_bot - 1] = 0
                self.BMat[row_idx_bot, row_idx_bot + self.Nx] = 0
                self.BMat[row_idx_bot, row_idx_bot - self.Nx] = 0

            # Y-boundaries only
            for image in range(1, self.Nx - 1):
                row_idx_top = image
                row_idx_bot = image + (self.Ny - 1) * self.Nx

                self.BMat[row_idx_top, row_idx_top] = 1
                self.BMat[row_idx_bot, row_idx_bot] = 1

                self.BMat[row_idx_top, row_idx_top + 1] = 0
                self.BMat[row_idx_top, row_idx_top - 1] = 0
                self.BMat[row_idx_top, row_idx_top + self.Nx] = 0

                self.BMat[row_idx_bot, row_idx_bot + 1] = 0
                self.BMat[row_idx_bot, row_idx_bot - 1] = 0
                self.BMat[row_idx_bot, row_idx_bot - self.Nx] = 0

            # Corner boundaries
            row_idx_00 = 0
            row_idx_N0 = self.Nx - 1
            row_idx_0N = (self.Ny - 1) * self.Nx
            row_idx_NN = (self.Ny * self.Nx) - 1

            self.BMat[row_idx_00, row_idx_00] = 1
            self.BMat[row_idx_00, row_idx_00 + 1] = 0
            self.BMat[row_idx_00, row_idx_00 + self.Nx] = 0

            self.BMat[row_idx_N0, row_idx_N0] = 1
            self.BMat[row_idx_N0, row_idx_N0 - 1] = 0
            self.BMat[row_idx_N0, row_idx_N0 + self.Nx] = 0

            self.BMat[row_idx_0N, row_idx_0N] = 1
            self.BMat[row_idx_0N, row_idx_0N + 1] = 0
            self.BMat[row_idx_0N, row_idx_0N - self.Nx] = 0

            self.BMat[row_idx_NN, row_idx_NN] = 1
            self.BMat[row_idx_NN, row_idx_NN - 1] = 0
            self.BMat[row_idx_NN, row_idx_NN - self.Nx] = 0

        elif self.BC == "open":
            for image in range(1, self.Ny):
                self.BMat[image * self.Nx - 1, image * self.Nx] = 0
                self.BMat[image * self.Nx, image * self.Nx - 1] = 0

        else:
            raise ValueError(
                f"Invalid boundary condition {self.BC}, cannot resolve "
                "diffusion matrix B"
            )

    def _get_col_idx(self, x_idx: int, y_idx: int) -> int:
        return x_idx + (y_idx * self.Ny)

    # ANCHOR need to update function signature
    def work_step(self):
        pass

    def flux_step(self):
        pass

    # TODO Migrate this to the base class
    def integrate_step(
        self, forceParams: Tuple, forceFunction_x: Callable,
        forceFunction_y: Callable
    ):
        if(self.split_method == 'lie'):
            self.advection_update(
                forceParams, forceFunction_x, forceFunction_y, self.dt
            )
            self.diffusion_update()

        elif(self.split_method == 'strang'):
            self.advection_update(
                forceParams, forceFunction_x, forceFunction_y, 0.5 * self.dt
            )
            self.diffusion_update()
            self.advection_update(
                forceParams, forceFunction_x, forceFunction_y, 0.5 * self.dt
            )

        elif(self.split_method == 'swss'):
            initProb = self.prob
            self.advection_update(
                forceParams, forceFunction_x, forceFunction_y, self.dt
            )
            self.diffusion_update()
            prob_1 = self.prob
            self.prob = initProb
            self.diffusion_update()
            self.advection_update(
                forceParams, forceFunction_x, forceFunction_y, self.dt
            )
            prob_2 = self.prob
            self.prob = 0.5*(prob_1 + prob_2)

        else:
            # Symmetric Strang splitting is the default choice
            self.advection_update(
                forceParams, forceFunction_x, forceFunction_y, 0.5 * self.dt
            )
            self.diffusion_update()
            self.advection_update(
                forceParams, forceFunction_x, forceFunction_y, 0.5 * self.dt
            )

    # TODO Keep this in the base class
    def diffusion_update(self):
        if (self.constDiff is True):
            self.prob = self.CMat.dot(self.prob)
        else:
            bVec = self.BMat.dot(self.prob)
            self.prob = scipy.sparse.linalg.spsolve(self.AMat, bVec)

    # TODO Also put this in the base class
    def advection_update(
        self, forceParams: Tuple, forceFunc_x: Callable, forceFunc_y: Callable,
        deltaT: float
    ):
        # if(self.adScheme=='lax-wendroff' or self.adScheme=='lw'):
        #    self.laxWendroff(forceParams,forceFunction,deltaT)
        # if(self.adScheme=='lax' or self.adScheme=='l'):
        #    self.lax(forceParams,forceFunction,deltaT)
        # if(self.adScheme=='lax-ds' or self.adScheme=='lds'):
        #    self.lax_dimSplit(forceParams,forceFunction,deltaT)
        # else:
        #   self.laxWendroff(forceParams,forceFunction,deltaT)
        #   self.lax_dimSplit(forceParams,forceFunction,deltaT)
        # self.laxWendroff_lieSplit_leg(forceFunc_x, forceFunc_y, forceParams, deltaT)
        pass

    # Probably just stick with LW for this? Just need to get it working...
    def lax_leg(self, forceParams: Tuple, forceFunction: Callable, deltaT: float):
        # TODO TEST LAX METHOD
        """
        This function updates the 2D probability density function using the 2-D lax-step method
        current implementation only uses PBCs
        """
        alpha = deltaT / (2 * self.dx)
        prob_mat = np.reshape(self.prob, (self.Nx, self.Ny))
        newProb = np.zeros_like(prob_mat)

        # First resolve 'bulk coordinates', '-2' counter means we want to evaluate boudary conditions separately
        for i in range(self.Nx - 2):
            for j in range(self.Ny - 2):
                # First calculate local average of NN probabilities
                newProb[i+1, j+1] = 0.25 * (
                    prob_mat[i+2, j+1] + prob_mat[i, j+1] + prob_mat[i+1, j+2] + prob_mat[i+1, j]
                )
                # Then calculate the forces
                newProb[i+1, j+1] -= alpha * (
                    forceFunction(self.xArray[i+2], self.yArray[j+1], forceParams, flag="x") * prob_mat[i+2, j+1] -
                    forceFunction(self.xArray[i], self.yArray[j+1], forceParams, flag="x") * prob_mat[i, j+1] +
                    forceFunction(self.xArray[i+1], self.yArray[j+2], forceParams, flag="y") * prob_mat[i+1, j+2] -
                    forceFunction(self.xArray[i+1], self.yArray[j], forceParams, flag="y") * prob_mat[i+1, j]
                )

        # Implement PBCs on boundary nodes
        # NOTE In this form, the 'corner boundaries' (where both x and y have simultaneous boudaries) are
        # double-counted, but this should only be ~2 lines per call
        for j in range(self.Ny):
            # Boundary conditions on 'x = 0' nodes
            newProb[0, j] = 0.25 * (
                prob_mat[1, j] + prob_mat[-1, j] + prob_mat[0, (j - 1) % self.Ny] + prob_mat[0, (j + 1) % self.Ny]
            )
            newProb[0, j] -= alpha * (
                forceFunction(self.xArray[1], self.yArray[j], forceParams, flag="x")
                * prob_mat[1, j]
                - forceFunction(self.xArray[-1], self.yArray[j], forceParams, flag="x")
                * prob_mat[-1, j]
                + forceFunction(self.xArray[0], self.yArray[(j + 1) % self.Ny], forceParams, flag="y")
                * prob_mat[0, (j + 1) % self.Ny]
                - forceFunction(self.xArray[0], self.yArray[(j - 1) % self.Ny], forceParams, flag="y")
                * prob_mat[0, (j - 1) % self.Ny]
            )

            # Boundary conditions on 'x = Nx' nodes
            newProb[-1, j] = 0.25 * (
                prob_mat[0, j] + prob_mat[-2, j] + prob_mat[-1, (j - 1) % self.Ny] + prob_mat[-1, (j + 1) % self.Ny]
            )
            newProb[-1, j] -= alpha * (
                forceFunction(self.xArray[0], self.yArray[j], forceParams, flag="x")
                * prob_mat[0, j]
                - forceFunction(self.xArray[-2], self.yArray[j], forceParams, flag="x")
                * prob_mat[-2, j]
                + forceFunction(self.xArray[-1], self.yArray[(j + 1) % self.Ny], forceParams, flag="y")
                * prob_mat[-1, (j + 1) % self.Ny]
                - forceFunction(self.xArray[-1], self.yArray[(j - 1) % self.Ny], forceParams, flag="y")
                * prob_mat[-1, (j - 1) % self.Ny]
            )

        for i in range(self.Nx):
            # Boundary conditions on 'y = 0' nodes
            newProb[i, 0] = 0.25 * (
                prob_mat[i, 1] + prob_mat[i, -1] + prob_mat[(i + 1) % self.Nx, 0] + prob_mat[(i - 1) % self.Nx, 0]
            )
            newProb[i, 0] -= alpha * (
                forceFunction(self.xArray[i], self.yArray[1], forceParams, flag="y")
                * prob_mat[i, 1]
                - forceFunction(self.xArray[i], self.yArray[-1], forceParams, flag="y")
                * prob_mat[i, -1]
                + forceFunction(self.xArray[(i + 1) % self.Nx], self.yArray[0], forceParams, flag="x")
                * prob_mat[(i + 1) % self.Nx, 0]
                - forceFunction(self.xArray[(i - 1) % self.Nx], self.yArray[0], forceParams, flag="x")
                * prob_mat[(i - 1) % self.Nx, 0]
            )

            # Boundary conditions on 'y = Ny' nodes
            newProb[i, -1] = 0.25 * (
                prob_mat[i, 0] + prob_mat[i, -2] + prob_mat[(i + 1) % self.Nx, -1] + prob_mat[(i - 1) % self.Nx, -1]
            )
            newProb[i, -1] -= alpha * (
                forceFunction(self.xArray[i], self.yArray[0], forceParams, flag="y")
                * prob_mat[i, 0]
                - forceFunction(self.xArray[i], self.yArray[-2], forceParams, flag="y")
                * prob_mat[i, -2]
                + forceFunction(self.xArray[(i + 1) % self.Nx], self.yArray[-1], forceParams, flag="x")
                * prob_mat[(i + 1) % self.Nx, -1]
                - forceFunction(self.yArray[(i - 1) % self.Nx], self.yArray[-1], forceParams, flag="x")
                * prob_mat[(i - 1) % self.Nx, -1]
            )

        # Flatten matrix probability and reform it into a single vector
        self.prob = np.reshape(newProb, (self.Nx * self.Ny))

    def lax_lieSplit_leg(
        self, forceFunc_x: Callable, forceFunc_y: Callable, forceParams: Tuple,
        deltaT: float
    ):
        """
        This function implements a Lax update of the 2D probability distribution based on a Lie dimensional splitting
        """
        # First cast the probabbility to a matrix shape
        prob_mat = np.reshape(self.prob, (self.Nx, self.Ny))
        newProb = np.zeros_like(prob_mat)

        # tempdt = self.dt
        self.dt = deltaT

        par_x = forceParams[0]
        par_y = forceParams[1]

        # Now, update the X dimension using the Lax scheme
        for i in range(self.Nx - 1):
            for j in range(self.Ny):
                newProb[i+1, j] = (
                    0.5 * (prob_mat[i + 2, j] + prob_mat[i, j])
                    - (self.dt / (2 * self.dx)) * (
                        forceFunc_x(self.xArray[i + 2], self.yArray[j], par_x)
                        * prob_mat[i + 2, j]
                        - forceFunc_x(self.xArray[i], self.yArray[j], par_x)
                        * prob_mat[i, j]
                    )
                )

        # TODO: Implement hard wall and open BCS here (and for y-coordinate below)
        # Resolve periodic boundary conditions along the x=0 and x=N boundaries
        for j in range(self.Ny):
            newProb[0, j] = (
                0.5 * (prob_mat[1, j] + prob_mat[-1, j])
                - (self.dt / (2 * self.dx)) * (
                    forceFunc_x(self.xArray[1], self.yArray[j], par_x)
                    * prob_mat[1, j]
                    - forceFunc_x(self.xArray[-1], self.yArray[j], par_x)
                    * prob_mat[-1, j]
                )
            )
            newProb[-1, j] = (
                0.5 * (prob_mat[0, j] + prob_mat[-2, j])
                - (self.dt / (2 * self.dx)) * (
                    forceFunc_x(self.xArray[0], self.yArray[j], par_x)
                    * prob_mat[0, j]
                    - forceFunc_x(self.xArray[-2], self.yArray[j], par_x)
                    * prob_mat[-2, j]
                )
            )

        # Now, update the y-coordinate using the Lax scheme
        for i in range(self.Nx):
            for j in range(self.Ny - 1):
                prob_mat[i, j+1] = (
                    0.5 * (newProb[i, j+2] + newProb[i, j])
                    - (self.dt / (2 * self.dy)) * (
                        forceFunc_y(self.xArray[i], self.yArray[j+2], par_y)
                        * newProb[i, j+2]
                        - forceFunc_y(self.xArray[i], self.yArray[j], par_y)
                        * newProb[i, j]
                    )
                )

        # Resolve periodic boundary conditions along the y=0 and y=N boundaries
        for i in range(self.Ny):
            prob_mat[i, 0] = (
                0.5 * (newProb[i, 1] + newProb[i, -1])
                - (self.dt / (2 * self.dy)) * (
                    forceFunc_y(self.xArray[i], self.yArray[1], par_y)
                    * newProb[i, 1]
                    - forceFunc_y(self.xArray[i], self.yArray[-1], par_y)
                    * self.newProb[i, -1]
                )
            )

            prob_mat[i, -1] = (
                0.5 * (newProb[i, 0] + newProb[i, -2])
                - (self.dt / (2 * self.dy)) * (
                    forceFunc_y(self.xArray[i], self.yArray[0], par_y)
                    * newProb[i, 0]
                    - forceFunc_y(self.xArray[i], self.yArray[-2], par_y)
                    * self.newProb[i, -2]
                )
            )

        # Update the probability vector
        self.prob = np.reshape(prob_mat, self.Nx * self.Ny)

        # tempdt = self.dt
        self.dt = deltaT

    def laxWendroff_lieSplit_leg(
        self, forceFunc_x: Callable, forceFunc_y: Callable, forceParams: Tuple,
        deltaT: float
    ):
        """
        This function performs an update on the advection term based on the
        Lax-Wendroff scheme for a 2D-diffusion, here the forceFunc_x and and
        ForceFunc_y are functions quantifying the force given a 2-D coordinate
        (x,y) and forceparams contains the individual parameters
        [[xParams],[yParams]] for the x and y force functions
        """

        # first shape probability into a matrix and declare two temporary matrices
        prob_mat = np.reshape(self.prob, (self.Nx, self.Ny))
        halfStep = np.zeros_like(prob_mat)
        probStar = np.zeros_like(prob_mat)

        tempdt = self.dt
        self.dt = deltaT

        par_x = forceParams[0]
        par_y = forceParams[1]

        # First calculate the x-update

        # Calculate half-step probabilities
        # NOTE: Updated modular calculation in arrays, should incorporate PBCS into this line
        # for i in range(self.Nx-1)
        for i in range(self.Nx):
            for j in range(self.Ny):
                halfStep[i, j] = (
                    0.5 * (prob_mat[(i + 1) % self.Nx, j] + prob_mat[i, j])
                    - (self.dt / (2 * self.dx)) * (
                        forceFunc_x(self.xArray[(i + 1) % self.Nx], self.yArray[j], par_x)
                        * prob_mat[(i + 1) % self.Nx, j]
                        - forceFunc_x(self.xArray[i], self.yArray[j], par_x)
                        * prob_mat[i, j]
                    )
                )

        # TODO: Implement hard wall and open boundary conditions for this
        # Resolve periodic boundaries for the half-step probabilities along the x=Nx boundary
        # for j in range(self.Ny):
        #    halfStep[-1,j] = 0.5*(prob_mat[0,j] + prob_mat[-1,j]) -
        # (self.dt/(2*self.dx))*(forceFunc_x(self.xArray[0],self.yArray[j],par_x)*prob_mat[0,j] -
        # forceFunc_x(self.xArray[-1],self.yArray[j],par_x)*prob_mat[-1,j])

        # Now update the probabilities in the x-direction based on the half-step fluxes
        # for i in range(self.Nx-1):
        for i in range(self.Nx):
            for j in range(self.Ny):
                probStar[(i + 1) % self.Nx, j] = (
                    prob_mat[(i + 1) % self.Nx, j]
                    - (self.dt / self.dx) * (
                        forceFunc_x(self.xArray[(i + 1) % self.Nx] + 0.5 * self.dx, self.yArray[j], par_x)
                        * halfStep[(i + 1) % self.Nx, j]
                        - forceFunc_x(self.xArray[i] + 0.5 * self.dx, self.yArray[j], par_x) * halfStep[i, j]
                    )
                )

        # And resolve the periodic boundary conditions on the x-updates
        # for j in range(self.Ny):
        #    probStar[0,j] = prob_mat[0,j] - (self.dt/self.dx)*(forceFunc_x(self.xArray[0]+
        # 0.5*self.dx,self.yArray[j],par_x)*halfStep[0,j] - forceFunc_x(self.xArray[-1]+0.5*self.dx,
        # self.yArray[j],par_x)*halfStep[-1,j])

        # Now perform the Y-updates on the probStar distribution
        # first calculate the half-step probabilities
        for i in range(self.Nx):
            # for j in range(self.Ny-1):
            for j in range(self.Ny):
                halfStep[i, j] = (
                    0.5 * (probStar[i, (j + 1) % self.Ny] + probStar[i, j])
                    - (self.dt / (2 * self.dy)) * (
                        forceFunc_y(self.xArray[i], self.yArray[(j + 1) % self.Ny], par_y)
                        * probStar[i, (j + 1) % self.Ny]
                        - forceFunc_y(self.xArray[i], self.yArray[j], par_y)
                        * probStar[i, j]
                    )
                )

        # Resolve the periodic boundaries for the half-step updates based on the y=Ny boundary
        # for i in range(self.Nx):
        #    halfStep[i,-1] = 0.5*(probStar[i,0] + probStar[i,-1]) - (self.dt/(2*self.dy))*
        # (forceFunc_y(self.xArray[i],self.yArray[0],par_y)*probStar[i,0] -
        # forceFunc_y(self.xArray[i],self.yArray[-1],par_y)*probStar[i,-1])

        # Now update the probability distribution based on the half-step fluxes
        for i in range(self.Nx):
            # for j in range(self.Ny-1):
            for j in range(self.Ny):
                prob_mat[i, (j + 1) % self.Ny] = (
                    probStar[i, (j + 1) % self.Ny]
                    - (self.dt / self.dy) * (
                        forceFunc_y(self.xArray[i], self.yArray[(j + 1) % self.Ny] + 0.5 * self.dy, par_y)
                        * halfStep[i, (j + 1) % self.Ny]
                        - forceFunc_y(self.xArray[i], self.yArray[j] + 0.5 * self.dy, par_y)
                        * halfStep[i, j]
                    )
                )

        # And resolve the boundary conditions for the full y-updates
        # for i in range(self.Nx):
        #    prob_mat[i,0] = probStar[i,0] - (self.dt/self.dy)*(forceFunc_y(self.xArray[i],
        # self.yArray[0]+0.5*self.dy,par_y)*halfStep[i,0] - forceFunc_y(self.xArray[i],
        # self.yArray[-1]+0.5*self.dy,par_y)*halfStep[i,-1])

        self.dt = tempdt
        # Update self.prob vector with result
        self.prob = np.reshape(prob_mat, self.Nx * self.Ny)

    # TODO write out hese routines
    def lax_wendroff(self):
        pass

    def _calcFlux_laxWendroff(self):
        pass

    def _getFluxDiff_laxWendroff(self):
        pass


if __name__ == "__main__":
    D = 1.0
    dt = 0.01
    dx = 0.1
    dy = 0.1
    xArray = np.arange(-1, 1, dx)
    yArray = np.arange(-1, 1, dy)

    fpe = FPE_integrator_2D(D, dt, dx, dy, xArray, yArray)


    fpe = FPE_integrator_2D(1.0, dt, dx, dy, xArray, yArray)




'''
Filename: Integrator.py

This python script contains the routines used to solve the Fokker-Planck
equation numerically via a split-integrator scheme with a 2-step Lax Wendroff
method for the forcing term and a semi-implicit Crank-Nicolson scheme for the
diffusion matrix

Author:         Steven Large
Created:        August 25th 2019

Version: 2.0.0

Software:       python 3.x
'''
from typing import Callable, Optional, Tuple
import numpy as np
from fokker_planck.base import Integrator
from fokker_planck.types.basetypes import (
    DiffScheme, BoundaryCondition, SplitMethod
)


class FokkerPlanck1D(Integrator):
    """1-dimensional Fokker-Planck integrator class, inherits from
    BaseIntegrator class
    """
    def __init__(
        self, D: float, dt: float, dx: float, x_array: np.ndarray,
        diff_scheme: DiffScheme = DiffScheme.CrankNicolson,
        boundary_cond: BoundaryCondition = BoundaryCondition.HardWall,
        split_method: SplitMethod = SplitMethod.Strang,
        output: bool = False,
        const_diffusion: bool = True
    ):
        """Constructor for 1-dimensional integrator object.

        Args:
            D (float): Diffusion coefficient for dynamics
            dt (float): Discrete time step size
            dx (float): Discrete spatial step size
            xArray (np.ndarray): Numpy array representing the discrete x-values
                of the system domain.
            diffScheme (Optional[str], optional): Diffusion scheme used for
                integration. Defaults to 'crank-nicolson'.
            adScheme (Optional[str], optional): Advection scheme used for
                integration. Defaults to 'lax-wendroff'.
            boundaryCond (Optional[str], optional): Boundary condition
                specification. Defaults to 'hard-wall'.
            splitMethod (Optional[str], optional): Operator splitting method
                used. Defaults to 'strang'.
            output (Optional[bool], optional): Flag for whether or not output
                is logged to console. Defaults to False.
            constDiff (Optional[bool], optional): Whether or not the diffusion
                term is constant in the integration course (D is unchanging).
                Defaults to True.
        """
        super().__init__(
            D, dt, diff_scheme, boundary_cond, split_method, output,
            const_diffusion
        )
        self.dx = dx

        self.N = len(x_array)
        self.prob = np.ones(self.N) / (self.N * self.dx)
        self.x_array = x_array

        self.init_diffusion_matrix()

    @property
    def dimension(self) -> int:
        return 1

    @property
    def mean(self) -> float:
        return np.sum(self.prob * self.x_array * self.dx)

    @property
    def variance(self) -> float:
        return np.sum(((self.x_array - self.mean) ** 2) * self.prob * self.dx)

    def reset(
        self, variance: Optional[float] = None, mean: Optional[float] = None
    ):
        """Routine to reset probability vector to unbiform, and re-initialize
        physical trackers, if BOTH variance and mean are provided, then the
        probability will be reinitialized to a Gaussian distribution with the
        input variance ane mean

        Args:
            variance (Optional[float], optional): Input varaince for Gaussian
                distribution. Defaults to None.
            mean (Optional[float], optional): Input mean for Gaussian
                distribution. Defaults to None.
        """
        if variance is not None and mean is not None:
            self.initialize_probability(mean, variance)
        else:
            self.prob = np.ones(self.N) / (self.N * self.dx)
        self.init_physical_trackers()

    def init_physical_trackers(self):
        """Routine to (re)initialize physical quantity trackers.
        """
        # Work and power tracking arrays
        self.workAccumulator = 0
        self.workTracker = []
        self.powerTracker = []
        self.timeTracker = []

        self.flux = np.zeros(len(self.x_array))
        # Total (integrated) flux tracker
        self.fluxTracker = 0

    def initialize_probability(self, mean: float, var: float):
        """Initialize 1-D Gaussian probability density

        Args:
            mean (float): mean of distribution
            var (float): variance of distribution
        """
        self.prob = np.exp(-(0.5 / var) * ((self.x_array - mean)**2))
        self.prob = self.prob / (sum(self.prob) * self.dx)

    def initialize_user_probability(
        self, func: Callable, params: Optional[Tuple] = None
    ):
        """Initlialize a user-provided probability density, based on input
        function `func` that has the signature func(x, *params)

        Args:
            func (Callable): Function with the signature (x, *params) that
                returns a probability as a function of x
            params (Optional[Tuple], optional): Tuple of parameters for the
                input probability function. Defaults to None.
        """
        self.prob = func(self.x_array, *params)
        self.prob = self.prob / (sum(self.prob) * self.dx)

    def check_CFL(self, force_params: Tuple, force_function: Callable) -> bool:
        """Routine to check whether or not Courant-Friedrichs-Lewy (CFL)
        criterion is satisfied for dynamics, given input force

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position

        Returns:
            bool: Flag as to whether or not the force function, parameters, and
                discretization evel satisfy the CFL
        """
        max_force = np.max(np.abs(force_function(self.x_array, force_params)))

        # set CFL and check if it is less than or greater than Unity
        self.CFL = max_force * self.dt / self.dx
        if (self.CFL > 1):
            if self.output:
                print("\t\tStability warning, invalid CFL --> " + str(self.CFL) + "\n\n")
            return False

        if self.output:
            print("\t\tCFL criterion satisfied, CFL --> " + str(self.CFL) + "\n\n")
        return True

    def init_diffusion_matrix(self):
        """Routine to initialize the A and B diffusion matrices for diffusion
        integration
        """
        # Set parameters for diffusion matrix iteration
        self._setDiffusionScheme()

        alpha = self.D * self.dt / (self.dx * self.dx)

        self.AMat = (
            np.diag(1 + 2 * alpha * self.exp_imp * np.ones(self.N))
            - np.diag(alpha * self.exp_imp * np.ones(self.N - 1), k=1)
            - np.diag(alpha * self.exp_imp * np.ones(self.N - 1), k=-1)
        )

        self.BMat = (
            np.diag(1 - 2 * alpha * (1 - self.exp_imp) * np.ones(self.N))
            + alpha * np.diag((1 - self.exp_imp) * np.ones(self.N - 1), k=1)
            + alpha * np.diag((1 - self.exp_imp) * np.ones(self.N - 1), k=-1)
        )

        # Initialize boundary columns based on self.BC
        self._initialize_boundary_terms(alpha)

        self.CMat = np.matmul(np.linalg.inv(self.AMat), self.BMat)

        # Test if sparse-matrix iteration steps are faster than normal matrix
        # multiplication
        self.test_sparse()

    def _initialize_boundary_terms(self, alpha: float):
        """Initialize boundary terms for diffusion matrices

        Args:
            alpha (float): _description_
        """
        # Left-side boundary
        self._matrix_boundary_A(alpha, 0)
        self._matrix_boundary_B(alpha, 0)

        # Right-side boundary
        self._matrix_boundary_A(alpha, self.N - 1)
        self._matrix_boundary_B(alpha, self.N - 1)

    def _matrix_boundary_A(self, alpha: float, idx: int):
        """Routine to set boudary-related terms in the diffusion matrix

        Args:
            alpha (float): coefficient for diffusion matrix terms, see
                documentation for definition
            idx (int): x-array index where

        Raises:
            ValueError: raised when self.BC parameter is invalid / not supported
        """
        if self.BC == BoundaryCondition.Periodic:
            self.AMat[idx, (idx + 1) % self.N] = -self.exp_imp * alpha
            self.AMat[idx, (idx - 1) % self.N] = -self.exp_imp * alpha

        elif self.BC == BoundaryCondition.HardWall:
            self.AMat[idx, idx] = 1 + 2 * alpha
            self.AMat[idx, abs(idx - 1)] = -2 * alpha

        elif self.BC == BoundaryCondition.Open:
            pass

        else:
            raise ValueError(
                f"Invalid boundary condition: {self.BC}, cannot resolve "
                "diffusion matrix A"
            )

    def _matrix_boundary_B(self, alpha: float, idx: int):
        """Determines / sets the parameters of matrix B on the boundaries

        Args:
            alpha (float): coefficient for diffusion matrix terms, see
                documentation for definition
            idx (int): x-array index where

        Raises:
            ValueError: raised when self.BC parameter is invalid / not supported
        """
        if self.BC == BoundaryCondition.Periodic:
            self.BMat[idx, (idx + 1) % self.N] = alpha * (1 - self.exp_imp)
            self.BMat[idx, (idx - 1) % self.N] = alpha * (1 - self.exp_imp)

        # NOTE double check this boundary resolution..
        elif self.BC == BoundaryCondition.HardWall:
            self.BMat[idx, idx] = 1
            self.BMat[idx, abs(idx - 1)] = 0

        elif self.BC == BoundaryCondition.Open:
            pass

        else:
            raise ValueError(
                f"Invalid boundary condition: {self.BC}, cannot resolve "
                "diffusion matrix A"
            )

    def work_step(
        self, force_params_pre: Tuple, force_params_post: Tuple,
        force_function: Callable, energy_function: Callable
    ):
        """Wrapper routine for integration step to calculate the work done due
        to changes in force function parameters. In this scheme, we effectively
        implement the Sekimoto definition of work, which splits work and heat
        into two steps: work occurs when system control parameters change,
        and dissipation follows afterwards. Thus, work is defined using an Ito
        discretization (changes occur at the start of the time step) while
        heat dissiaption occurs afterwards.  Thus, the force parameters here
        are used to calculate work, and then the system responds to the values
        of the force based on `forceParams_post`.

        Args:
            forceParams_pre (Tuple): Force parameters from before (pre) update
            forceParams_post (Tuple): Force parameters form after (post) update
            forceFunction (Callable): Force function
            energyFunction (Callable): Energy function that the forceFunction
                is derived from 
        """

        # Calculate average energy before and after update to force parameters
        currEnergy = (
            sum(energy_function(self.x_array, force_params_pre) * self.prob) * self.dx
        )
        newEnergy = (
            sum(energy_function(self.x_array, force_params_post) * self.prob) * self.dx
        )

        self.integrate_step(force_params_post, force_function)

        # Work done per step is given by the change in average energy through
        # force function update
        self.work_accumulator += newEnergy - currEnergy
        self.work_tracker.append(self.work_accumulator)
        # Power is the change in work, divided by the time over which the change
        # occurred (which is dt here)
        self.power_tracker.append((newEnergy - currEnergy) / self.dt)

    def flux_step(self, force_params: Tuple, force_function: Callable):
        """Similar to work_step(...) this routine wraps the integrate_step in
        the parent class with calcualtions for the probability flux. Here
        the flux J is defined in the alternate version of the FPE:

            partial_t p = - partial_x J

        And so the flux can be calculated relatively simply by te current
        values of the probability, as well as the force parameters

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
        """
        # flux as a function of position
        self.flux = (
            self.D * force_function(self.x_array, force_params) * self.prob
            - self.D * np.gradient(self.prob)
        )
        # Calculate integrated (net) flux over current configuration
        self.flux_tracker = sum(self.flux) * self.dx
        self.integrate_step(force_params, force_function)

    def lax_wendroff(
        self, force_params: Tuple, force_function: Callable, delta_t: float
    ):
        """Implementation of lax-wendroff method for 1-D system

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
            deltaT (float): Size of time step (which is not necessarily equal
                to the self.dt parameter because of the operator splittings)
        """
        # Initialize empty array for new probability
        new_prob = np.zeros(len(self.prob))

        # Get half-time-step flux terms
        halfFlux = self._calc_flux_lax_wendroff(
            force_params, force_function, delta_t
        )

        # Update probability with Lax-step using half-time-step flux terms
        for index in range(len(self.prob)):
            new_prob[index] = (
                self.prob[index]
                - (self.D * delta_t / self.dx)
                * (halfFlux[index+1] - halfFlux[index])
            )

        # set probability to be update prob
        self.prob = new_prob

    def _calc_flux_lax_wendroff(
        self, force_params: Tuple, force_function: Callable, delta_t: float
    ) -> np.ndarray:
        """Routine to calculate the half-time-step flux terms for the
        lax-wendroff advection update scheme

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
            deltaT (float): size of temporal discretization (dt) for
                integration step

        Returns:
            np.ndarray: half-time-step flux terms, at positions x_min = -1/2 dx
            up to x_max = (N + 1/2) dx
        """

        # initialize half-step probability and half-step flux terms
        half_prob = np.zeros(len(self.prob) + 1)
        half_flux = np.zeros(len(self.prob) + 1)

        # Bulk terms
        for i in range(len(self.prob) - 1):

            half_prob[i + 1] = (
                0.5 * (self.prob[i + 1] + self.prob[i])
                - self._get_flux_diff_lax_wendroff(force_function, force_params, delta_t, i)
            )
            half_flux[i + 1] = (
                force_function(self.x_array[i] + 0.5 * self.dx, force_params) * half_prob[i + 1]
            )

        # Boundary terms
        if self.BC == BoundaryCondition.Periodic:
            half_prob[0] = (
                0.5 * (self.prob[0] + self.prob[-1])
                - self._get_flux_diff_lax_wendroff(
                    force_function, force_params, delta_t, len(self.prob) - 1
                )
            )

            half_prob[-1] = (
                0.5 * (self.prob[-1] + self.prob[0])
                - self._get_flux_diff_lax_wendroff(force_function, force_params, delta_t, 0)
            )
 
            half_flux[0] = (
                force_function(self.x_array[0] - 0.5 * self.dx, force_params) * half_prob[0]
            )
            half_flux[-1] = half_flux[0]

        elif self.BC == BoundaryCondition.Open:

            fluxFw = (
                (self.D * delta_t / (2 * self.dx))
                * force_function(self.x_array[0], force_params)
                * self.prob[0]
            )
            fluxRev = (
                (self.D * delta_t / (2 * self.dx))
                * force_function(self.x_array[-1], force_params)
                * self.prob[-1]
            )
            half_prob[0] = 0.5 * self.prob[0] - fluxFw
            half_prob[-1] = 0.5 * self.prob[-1] + fluxRev
            half_flux[0] = force_function(self.x_array[0] - 0.5 * self.dx, force_params) * half_prob[0]
            half_flux[-1] = force_function(self.x_array[-1] + 0.5 * self.dx, force_params) * half_prob[-1]

        else:
            # Hard wall boundaries
            half_prob[0] = 0.5 * self.prob[0]
            half_prob[-1] = 0.5 * self.prob[-1]
            half_flux[0] = force_function(self.x_array[0] - 0.5 * self.dx, force_params) * half_prob[0]
            half_flux[-1] = force_function(self.x_array[-1] + 0.5 * self.dx, force_params) * half_prob[-1]

        return half_flux

    def _get_flux_diff_lax_wendroff(
        self, force_function: Callable, force_params: Tuple, delta_t: float,
        idx: int
    ) -> float:
        """Calculate the difference between forward and reverse flux terms at
        half-step points within the LW method

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
            deltaT (float): Discrete time step size for integration step
            idx (int): Index for half-step update (which is i - 1/2 in terms of
                the original spatial mesh)

        Returns:
            float: Flux difference between i-1/2 and i+1/2 in spatial
                discretization
        """
        fluxFw = (
            (self.D * delta_t / (2 * self.dx))
            * force_function(self.x_array[(idx + 1) % len(self.x_array)], force_params)
            * self.prob[(idx + 1) % len(self.x_array)]
        )
        fluxRev = (
            (self.D * delta_t / (2 * self.dx))
            * force_function(self.x_array[idx], force_params)
            * self.prob[idx]
        )

        return fluxFw - fluxRev


if __name__ == "__main__":
    D = 1
    dx = 0.01
    dt = 0.01
    x_array = np.arange(-1, 1, dx)
    fpe = FokkerPlanck1D(D, dt, dx, x_array)

    print("-- DONE --")



# Base class for FPE integrator

from abc import ABC, abstractmethod
from typing import Callable, Tuple
import numpy as np
import time
import scipy.sparse

from fokker_planck.types.basetypes import DiffScheme, BoundaryCondition, SplitMethod


class Integrator(ABC):
    """Base class for integrator objects (abstract), all instances of FPE
    integrators must inherit from this base class. Universal functions are
    defined and implemented here, while instance-specific methods are
    relegated to the inherited classes

    Raises:
        ValueError: _description_
        NotImplementedError: _description_
        ValueError: _description_
        ValueError: _description_
    """
    def __init__(
        self, D: float, dt: float, diff_scheme: DiffScheme,
        boundary_cond: BoundaryCondition, split_method: SplitMethod,
        output: bool = False, const_diffusion: bool = True
    ):
        """Constructor for base integrator class

        Args:
            D (float): diffusion coefficient for dynaimcs (assumed to be constant)
            dt (float): Discrete ime step size
            diffScheme (str): Integration scehem used for diffusion step: can
                be one of `explicit`, `implicit` or `crank-nicolson`, default
                is to use `crank-nicolson`
            adScheme (str): INtegration scheme used for advection step, for now
                this is a passthrough method as only the 2-step lax-wendroff
                method is implemented
            boundaryCond (str): Boundary condition to use on the domain, can be
                one of `hard-wall`, `periodic`, or 'open', default is to use
                `hard-wall`
            splitMethod (str): Operator splitting method used to segment the 
                advection and diffusiojn integration steps. Default is symmetric
                Strang splitting
            output (Optional[bool], optional): Whether to show internal output
                messages during execution. Defaults to False.
            constDiff (Optional[bool], optional): Whether or not the diffusion
                coefficient will be constant through the course of simulation.
                Defaults to True.
        """
        # Validate input parameters
        self._validate_input(
            diff_scheme, boundary_cond, split_method
        )

        # Initialize instance variables. For now, this routine supports
        # constant diffusion coefficient integrations (i.e. D is not a tensor)
        self.D = D
        self.dt = dt

        # Condition for satisfying CFL criterion
        self.CFL = None

        # Instantiate the instance directives
        self.diff_scheme = diff_scheme    # Integration scheme for diffusion term
        self.BC = boundary_cond          # Boundary conditions
        self.split_method = split_method  # Integator splitting method
        self.const_diffusion = const_diffusion      # Flag for if diffusion matrix is const

        # Flag for whether or not output will be printed
        self.output = output

        # Default initial argument for sparse method utilization in matrix
        # operations
        self.sparse_calc = False

        # Physics trackers
        self.work_accumulator = 0
        self.work_tracker = []
        self.power_tracker = []

    def _validate_input(
        self, diff_scheme: DiffScheme, boundary_cond: BoundaryCondition,
        split_method: SplitMethod
    ):
        """Validation method to handle input parameters. In this method the
        checks are for whether or not the input diffusion scheme, advection
        scheme and boundary conditions are supported, by seeing if the keys
        are present in the config.py file

        Args:
            diffScheme (str): Diffusion scheme identifier
            adScheme (str): Advection scheme identifier
            boundaryCond (str): Boundary conditions
            splitMethod (str): Operator splitting method

        Raises:
            ValueError: Diffusion scheme not recognized
            NotImplementedError: Advection scheme not recognized
            ValueError: boundary condition not recognized
            ValueError: Operator splitting scheme not recognized
        """
        if not isinstance(diff_scheme, DiffScheme):
            raise ValueError("diffScheme not recognized, see config file")

        if not isinstance(boundary_cond, BoundaryCondition):
            raise ValueError(
                "boundary condition not recognized, see config file"
            )

        if not isinstance(split_method, SplitMethod):
            raise ValueError(
                "Splitting method not recognized, see config file"
            )

    def _setDiffusionScheme(self):
        """Method to set the diffusion scheme in the integrator instance. In
        all cases there are three options: implicit, explicit, and
        crank-nicolson. The latter is the default choice.
        """
        if (self.diff_scheme == DiffScheme.Implicit):
            self.exp_imp = 1.0

        elif (self.diff_scheme == DiffScheme.Explicit):
            self.exp_imp = 0.0

        elif (self.diff_scheme == DiffScheme.CrankNicolson):
            self.exp_imp = 0.5

        else:
            if (self.output):
                print("\t\tInt scheme unknown, using default setting (CN)...")
            self.exp_imp = 0.5

    def test_sparse(self):
        """Routine to check whether sparse matrix calculations are more rapid
        than standard matrix multiplication techniques. This is expected to be
        the case for large matrices, and more impactful when diffisuon is not
        constant. For constant diffusion term, the update due to diffusion is
        only a matrix multiplication (and not a matrix inversion) so the
        scaling is better with N.
        """
        if (self.const_diffusion):
            # NOTE Implement checks here
            self.spase_calc = False

        else:
            sBMat = scipy.sparse.csr_matrix(self.BMat)
            sAMat = scipy.sparse.csr_matrix(self.AMat)
            self._test_sparse(sAMat, sBMat)

    def _test_sparse(
        self, sAMat: scipy.sparse.csr_matrix,
        sBMat: scipy.sparse.csr_matrix
    ):
        """Method to check whether full diffusion update (matrixc inversion) is
        faster when sparse matrix data structures are used, as opposed to
        standard matrix formats.

        Args:
            sAMat (scipy.sparse.csr_matrix): Sparse formatted verison of
                instance variable AMat
            sBMat (scipy.sparse.csr_matrix): Sparse matrix formatted version of
                instance variable BMat
        """
        # Start timer
        start_sparse_full = time.time()
        sbVec = sBMat.dot(self.prob)
        _ = scipy.sparse.linalg.spsolve(sAMat, sbVec)
        # Stop timer
        end_sparse_full = time.time()
        time_sparse = end_sparse_full - start_sparse_full

        # Start timer (non-sparse calculations)
        start_reg_full = time.time()
        bVec = np.matmul(self.BMat, self.prob)
        _ = np.linalg.solve(self.AMat, bVec)
        # stop timer
        end_reg_full = time.time()
        time_reg = end_reg_full - start_reg_full

        # If sparse calculation is faster, use sparse matrix methods
        if (time_sparse < time_reg):
            if (self.output):
                print("\t\tSparse matrix methods preferred...")
            self.sparse_calc = True
            self.BMat = sBMat
            self.AMat = sAMat

        # Otherwise, use standard dense matrix calculation techniques
        else:
            if (self.output):
                print("\t\tDense matrix methods preferred...")
            self.sparse_calc = False

    def integrate_step_advection(
        self, force_params: Tuple, force_function: Callable
    ):
        """Method to indtegrate a step of ONLY the advection term (so no
        diffusion). This cannot be done, simply, by setting D = 0 as terms in
        the advection equation depend on D (because the dynamics of the system
        are governed by D in the FPE).

        Args:
            forceParams (Tuple): Parameters for the force function
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
        """
        self.advection_update(force_params, force_function, self.dt)

    def diffusion_update(self):
        """Update to the system probability based on the diffusion dynamics
        """
        if (self.sparse_calc):
            bVec = self.BMat.dot(self.prob)
            self.prob = scipy.sparse.linalg.spsolve(self.AMat, bVec)
        else:
            if(self.const_diffusion):
                self.prob = np.matmul(self.CMat, self.prob)
            else:
                bVec = np.matmul(self.BMat, self.prob)
                self.prob = np.linalg.solve(self.AMat, bVec)

    def advection_update(
        self, force_params: Tuple, force_function: Callable, delta_t: float
    ):
        """Update the advective component of the equation

        Args:
            forceParams (Tuple): Parameters for force function
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
            deltaT (float): Time step for dynamics
        """
        # Currently, this only directs towards the LW routine, but supports
        # the flexibility to add others in the future
        self.lax_wendroff(force_params, force_function, delta_t)

    def integrate_step(self, force_params: Tuple, force_function: Callable):
        """Integrates the equations by one time step. In this package, operator
        splitting methods are used, and this interface routine directs the update
        step towards the correct endpoint: Strang, Lie, ot SWSS

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
        """
        if (self.split_method is SplitMethod.Lie):
            self._integrate_step_lie(force_params, force_function)

        elif (self.split_method is SplitMethod.Strang):
            self._integrate_step_strang(force_params, force_function)

        elif (self.split_method is SplitMethod.SymStrang):
            self._integrate_step_SWSS(force_params, force_function)
        else:
            self._integrate_step_strang(force_params, force_function)

    def _integrate_step_lie(self, force_params: Tuple, force_function: Callable):
        """Implementation of Lie splitting update of FPE

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
        """
        self.advection_update(force_params, force_function, self.dt)
        self.diffusion_update()

    def _integrate_step_strang(
        self, force_params: Tuple, force_function: Callable
    ):
        """Implementation of (symmetric) Strang Splitting update of FPE

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
        """
        self.advection_update(force_params, force_function, 0.5 * self.dt)
        self.diffusion_update()
        self.advection_update(force_params, force_function, 0.5 * self.dt)

    def _integrate_step_SWSS(
        self, force_params: Tuple, force_function: Callable
    ):
        """Implementation of the Symmetrically weighted Strang splitting update
        of the FRE.

        Args:
            forceParams (Tuple): Tuple representing the current force parameters
            forceFunction (Callable): The function taking arguments of the form
                (x, *args) that gives the force on the system as function of
                its position
        """
        initProb = self.prob
        self.advection_update(force_params, force_function, self.dt)
        self.diffusion_update()
        prob_1 = self.prob
        self.prob = initProb
        self.diffusion_update()
        self.advection_update(force_params, force_function, self.dt)
        prob_2 = self.prob
        self.prob = 0.5 * (prob_1 + prob_2)

    @property
    def get_prob(self) -> np.ndarray:
        return self.prob

    @abstractmethod
    def init_physical_trackers(self):
        pass

    @abstractmethod
    def init_diffusion_matrix(self):
        pass

    @abstractmethod
    def check_CFL(self, force_params: Tuple, force_function: Callable) -> bool:
        pass

    @abstractmethod
    def lax_wendroff(self, *args, **kwargs):
        pass



# this python module contains all of the force functions for the FPE integrator
#
# Steven Large
# August 26th 2019

# import os
import numpy as np
import sys
from typing import Tuple, Union, Iterable

InputParameters = Union[Tuple, Iterable]
PositionData = Union[float, np.ndarray]

ForceData = Union[float, np.ndarray]
EnergyData = Union[float, np.ndarray]


def no_force(
    position: PositionData, params: InputParameters
) -> ForceData:
    """Null force function, returns zero always

    Args:
        position (float ot numpy array): System position
        params (tuple): Parameters

    Returns:
        ForceData: zero force
    """
    return 0


def constant_force(
    position: PositionData, params: InputParameters
) -> ForceData:
    """Constant force function, force parameters are of form [kValue]
    where kValue is the force.

    Potential energy is -p_0 * position

    p_0 = Constant force value

    Args:
        position (PositionData): system position
        params (ForceParameters): constant force value

    Returns:
        ForceData: force at each position point
    """
    return params[0]


def harmonic_force(
    position: PositionData, params: InputParameters
) -> ForceData:
    """Harmonic potential force, with potential of the form

    E(x) = p_0 / 2 * (position - p_1)^2

    p_0 = spring constant
    p_1 = potential energy minimum position

    Args:
        position (PositionData): system position
        params (ForceParameters): poatential parameters

    Returns:
        ForceData: forces
    """
    return -params[0] * (position - params[1])


def harmonic_energy(
    position: PositionData, params: InputParameters
) -> EnergyData:
    """Harmonic energy function

    E(x) = p_0 * (x - p_1)^2

    p_0 = spring constant
    p_1 = potential energy minimum location

    Args:
        position (PositionData): system position
        params (InputParameters): Energy parameters

    Returns:
        EnergyData: Harmonic energy at the input position
    """
    return 0.5 * params[0] * ((position - params[1])**2)


def harmonic_force_const_velocity(
    position: PositionData, params: InputParameters
) -> ForceData:
    """Effecive force function for a harmonic trap with its minimum moving at
    a constant velocity. In this potential it is assmed that the x-coordinate
    reference frame situates the trap minimum at x=0 (the refernece frame co-moves
    with the harmonic potential). (beta = 1/kT is assumed to be unity)

    The force function is of the form:

    F(x) = -p_0 * x - p_1 / p_2

    p_0 = Spring constant
    p_1 = Trap center (assumed to be zero for forces)
    p_2 = Trap velocity
    p_3 = Diffusion coefficient

    Args:
        position (PositionData): system posision
        params (InputParameters): force function parameters

    Returns:
        ForceData: Force at each input position
    """
    return -params[0] * (position - params[1]) - params[2] / params[3]


def harmonic_energy_const_velocity(
    position: PositionData, params: InputParameters
) -> EnergyData:
    """Energy function corresponding to a centered harmonic potential, but with
    a parameter signature that matches the harmonic force constant-velocity
    force function

    Args:
        position (PositionData): system posision
        params (InputParameters): force function parameters

    Returns:
        EnergyData: Energy at each input position
    """
    return 0.5 * params[0] * (position - params[1]) ** 2


def periodic_force(
    position: PositionData, params: InputParameters
) -> ForceData:
    """Force produced by a sinusoindal, periodic potential.  Force function is
    of the form

    F(x) = - (pi * p_0 / p_1) * sin((2 * pi / p_1) * x - p_2 )

    p_0: Barrier height in sinusoidal potential
    p_1: Number of minima/maxima per rotation
    p_2: Phase shift (in rotations)

    Args:
        position (PositionData): System position (in rotations: 0 - 1
            corresponds to a full rotation)
        params (InputParameters): Input force parameters

    Returns:
        ForceData: Force as a function of the input rotation value(s)
    """
    force = (
        -(np.pi * params[0] * params[1])
        * np.sin((2 * np.pi * params[1]) * (position - params[2]))
    )
    return force


def periodic_energy(
    position: PositionData, params: InputParameters
) -> EnergyData:
    """Energy of a periodic potential energy function, taking the form

    E(x) = -(p_0 / 2) * cos(2 * pi * p_1 * (x - p_2))

    p_0: Height of energy barrier of potential (in units of beta = 1 / kT)
    p_1: Number of minima and maxima in a single rotation
    p_2: Phase shift, in rotations

    Args:
        position (PositionData): System position
        params (InputParameters): Energy function parameters

    Returns:
        EnergyData: energy as a function of rotation
    """
    energy = -0.5 * params[0] * np.cos((2 * np.pi * params[1]) * (position - params[2]))
    return energy


def periodic_force_const_velocity(
    position: PositionData, params: InputParameters
) -> ForceData:
    # TODO There should probably be a diffusion coefficient in here?
    """Force applied to a periodic system when the potential is rotating at a
    constant velocity. Force function is of the form:

    F(x) = - p_3 - pi * p_0 * p_1 * sin(2 * pi * p_1 * (x - p_2))

    p_0: Height of energy barrier of potential (in units of beta = 1 / kT)
    p_1: Number of minima and maxima in a single rotation
    p_2: Phase shift, in rotations
    p_3: Velocity of rotation, in rotations / unit time

    Args:
        position (PositionData): System position
        params (InputParameters): Input force parameters

    Returns:
        ForceData: Force at each of the input position(s)
    """
    return -params[2] - np.pi * params[0] * np.sin(2 * np.pi * (position - params[1]))


def bistable_force(
    position: PositionData, params: InputParameters
) -> ForceData:
    """Simple potential energy function for a bistable potential of the form.
    The energy potential is bistable when p_0 > 0, p_1 > 0 so the parameters
    are restricted to be in that range

    F(x) = p_0 * x - p_1 * x^3

    p_0: strength of parabolic energy contribution
    p_1: strength of quartic energy contribution

    Args:
        position (PositionData): System position
        params (InputParameters): Input force parameters

    Returns:
        ForceData: Force as a function of position
    """
    return params[0] * (position) - params[1] * (position**3)


def bistable_energy(
    position: PositionData, params: InputParameters
) -> EnergyData:
    """Energy function for simple bistable potential of the form:

    E(x) = -p_0/2 * x^2 + p_1/4 * x^4

    p_0: Strength of quadratic potential
    p_1: Strength of quartic potential

    Args:
        position (PositionData): System postiion
        params (InputParameters): Input energy parameters

    Returns:
        EnergyData: Energy as a function of position
    """
    return -0.5 * params[0] * (position ** 2) + 0.25 * params[1] * (position ** 4)


def ABEL_force(position: PositionData, params: InputParameters) -> ForceData:
    """Force produced by a bistable ABEL trap from Ref[1]. This
    function is parameterized so that there will be potential energy minima
    at x = +- 1. The energy function is of the form

    E(x) = 4 p_0 * (1/4 * x^4 - p_1/2 * x^2 - p_2 * p_3 * x)

    p_0: Height of energy barrier separating the two minima (when p_1 = 1)
    p_1: Control parameter dictating the height of the separating barrier
        (p_1 is in [0, 1])
    p_2: Control parameter indicating the strength of a linear perturbation to
        the potential, biasing the system into one of the two wells (p_2 is in [0, 1])
    p_3 (Optional): Parameter indicating the degree of erasure, default is 0.5
        (full erasure) but can be adjusted to larger values as well. This parameter
        is not considered a dynamic control parameter, but simply defines the
        scale of the linear term energetic bias.

    [1] Y. Jun, M. Gavrilov & J. Bechhoefer "High-precision test of Landauer's
    Principle in a Feedback Trap", Phys. Rev. Lett., 2014, 113, 190601

    Args:
        position (PositionData): Position of the system
        params (InputParameters): Energy parameters

    Returns:
        ForceData: Force as a function of the input positions

    """
    if len(params) > 3:
        A = params[3]
    else:
        A = 0.5
    return -4 * params[0] * (position ** 3 - params[1] * position - A * params[2])


def ABEL_energy(position: PositionData, params: InputParameters) -> EnergyData:
    """Potential energy function of a bistable ABEL trap from Ref[1]. This
    function is parameterized so that there will be potential energy minima
    at x = +- 1. The energy function is of the form

    E(x) = 4 p_0 * (1/4 * x^4 - p_1/2 * x^2 - p_2 * p_3 * x)

    p_0: Height of energy barrier separating the two minima (when p_1 = 1)
    p_1: Control parameter dictating the height of the separating barrier
        (p_1 is in [0, 1])
    p_2: Control parameter indicating the strength of a linear perturbation to
        the potential, biasing the system into one of the two wells (p_2 is in [0, 1])
    p_3 (Optional): Parameter indicating the degree of erasure, default is 0.5
        (full erasure) but can be adjusted to larger values as well. This parameter
        is not considered a dynamic control parameter, but simply defines the
        scale of the linear term energetic bias.

    [1] Y. Jun, M. Gavrilov & J. Bechhoefer "High-precision test of Landauer's
    Principle in a Feedback Trap", Phys. Rev. Lett., 2014, 113, 190601

    Args:
        position (PositionData): Position of the system
        params (InputParameters): Energy parameters

    Returns:
        EnergyData: Energy as a function of the input positions
    """
    if len(params) > 3:
        A = params[3]
    else:
        A = 0.5

    return 4 * params[0] * (0.25 * (position ** 4) - 0.5 * params[1] * (position ** 2) - A * params[2] * position)


def ABEL_force_v2(position, params):
    '''
    ABEL force v2 (current version, circa 2019), force parameters: [a,xm]
    potential energy:   0.5*a*(x+xm)^2    IF  x < -xp
                        -0.5*(2*eDag/(xm^2 - 2*eDag/a))x^2 + eDag   IF -xp < x < xp
                        0.5*a*(x-xm)^2    IF  x > xp
    '''
    pass


def ABEL_energy_v2(position, params):
    pass


'''
Hairpin forces, force parameters?
'''

'''
KSS Model potential (eriodic version of it...)
'''


'''
# ANCHOR ------ MULTIDIMENSIONAL POTENTIALS ------
'''


def const_force(
    x: PositionData, y: PositionData, params: InputParameters
) -> ForceData:
    """Constant force funtion for 2-dimensional potential energy function,
    assumes that the constant foce values is the same in both X and Y
    directions. Force is if the function

    F(x, y) = p_0

    Args:
        x (PositionData): x-coordinate data
        y (PositionData): y-coordinate data
        params (InputParameters): input parameters

    Returns:
        ForceData: Constant force value
    """
    return params[0]


# ANCHOR Harmonic potential
def harmonic_energy_2D(
    x: PositionData, y: PositionData, params: InputParameters
) -> EnergyData:
    """2-dimensional harmonic energy function of the form:

    E(x, y) = p_00 / 2 * (x - p_01)^2 + p_10 / 2 * (y - p_11)^2

    Parameters are arranged in an array, where the first index determines the
    dimension they act on (x or y) and the second index indicates the parameter
    within that direction

    p_00: Spring constant in x-coordinate
    p_01: Trap center in x-coordinate
    p_10: Spring constant in y-coordinate
    p_11: Trap center in y-coordinate

    Args:
        x (PositionData): System x-position values
        y (PositionData): SyStem y-position values
        params (InputParameters): Input energy parameters

    Returns:
        EnergyData: energy as a function of the input positions
    """
    return 0.5 * (params[0][0] * ((x - params[0][1]) ** 2) + params[1][0] * ((y - params[1][1]) ** 2))


def harmonic_force_z(
    z: PositionData, params: InputParameters
) -> ForceData:
    """Force in the z-direction (where z is x or y) due to 2-D harmonic
    potential energy function. For this function, we only need parameters from
    the z-direction. The energy function is of
    the form

    E(x, y) = p_00/2 * (x - p_01)^2 + p_10/2 * (x - p_11)^2

    So that (for z = x)

    F_x(x, y) = -par_x E(x, y) = -p_00 * (x - p_01)

    p_0: Spring constant in z-direction
    p_1: Trap center in z-direction

    Args:
        z (PositionData): z-coordinate position
        params (InputParameters): Parameters for z-direction dynamics

    Returns:
        ForceData: Force along the z-direction for input z data
    """
    return -params[0] * (z - params[1])


# ANCHOR Periodic potential
def periodic_energy_2D(
    x: PositionData, y: PositionData, params: InputParameters
) -> EnergyData:
    """2-D periodic energy function, given by the equation

    E(x, y) = p_00 / 2 * cos(2 * pi * p_01 * x) + p_10 / 2 * cos(2 * pi * p_11 * y)

    p_00: Energy barrier height along x-direction
    p_01: Number of potential energy minima and maxima per rotation in x
    p_10: Energy barrier height along the y-direfction
    P_11: Number of potential energy minima and maxima per rotation in y

    Args:
        x (PositionData): x position coordimates
        y (PositionData): y position coordinates
        params (InputParameters): Parameters defining the X-Y potential energy function

    Returns:
        EnergyData: Potential energy as a function of the x,y coordaintes input
    """
    E_x = 0.5 * params[0][0] * np.cos(2 * np.pi * params[0][1] * x)
    E_y = 0.5 * params[1][0] * np.cos(2 * np.pi * params[1][1] * y)
    return E_x + E_y


def gen_periodic_landscape_2D(x_array, y_array, params):
    energy = periodic_energy_2D(x_array, y_array, params)
    force = np.gradient(energy)
    force_x, force_y = force[0], force[1]
    return energy, force_x, force_y


def periodic_force_2D(x_val, y_val, params, flag="z"):
    if flag == "x":
        return np.pi * params[0] * np.sin(2 * np.pi * x_val)
    if flag == "y":
        return np.pi * params[1] * np.sin(2 * np.pi * y_val)
    else:
        print("Error: unidentified force flag value, exiting program...")
        sys.exit()



from setuptools import find_packages, setup

# README contents
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='fokker-planck',
    packages=find_packages(),
    version='1.0.2',
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='''
        A Fokker-Planck integrator in python
    ''',
    author='Steven Large',
    author_email='stevelarge7@gmail.com',
    license='MIT',
    install_requires=[

    ],
    project_urls={
        'Source': 'https://github.com/StevenJLarge/FokkerPlanck'
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Mathematics"
    ]

)
